/*
** Copyright 1995-1996 by the Massachusetts Institute of Technology.  All
** rights reserved. For more information, see the file COPYING which came
** with this source distribution.
*/

#ifndef MIDIMSG_H
#define MIDIMSG_H

#include "basic_types.h"
#include "UtilString.h"
#include "UtilPtr.h"
#include "UtilArray.h"

//--------------------------------------------------------------------------------------------------

struct smsgt
{
	enum
	{
   	// System Exclusive message
   	sys_ex              =   0,
   	
		// System Common messages
		mtc                 =   1, 
		song_position       =   2,
		song_select         =   3,
		sys_com_undefined_1 =   4, ext = sys_com_undefined_1,
		sys_com_undefined_2 =   5,
		tune_request        =   6,
		end_sys_ex          =   7, file_escape = 7, // 1
   	
		// System Realtime messages
		timing_clock        =   8,  
		sys_rea_undefined_1 =   9,
		start               = 0xA,
		cont                = 0xB,
		stop                = 0xC,
		sys_rea_undefined_2 = 0xD,
		active_sensing      = 0xE,
		system_reset        = 0xF, file_meta = 0xF
	};
};

//--------------------------------------------------------------------------------------------------


struct baws // byte array with size
{
	typedef UtilArray<uchar> ucharArray;
	UtilPtr<ucharArray> d;
	
	baws() { };
	baws(const baws& b) { operator=(b); }
	baws(const uchar* r, long s) { init(r, s); }
	void init( const uchar *r, long s );
	baws& operator=(const baws& b)
	{
		d = b.d;
		return (*this);
	}
};

//--------------------------------------------------------------------------------------------------
class
meta
{	
private:
	static uint2 swap2(uint2 n)
	{
		unsigned long test = 0x12345678;
		if (*((unsigned char*)(&test)) == 0x12)
			return n;
		else
			return ((n >> 8) | ((n & 0xff) << 8));
	}

public:
	struct offs
	{
		enum fmt { f24, f25, f30d, f30 } format;
		uchar hr, min, sec, fr, centifr;
	};
	struct tsig { uchar num, denom, Mclks_p_beat, not32s_p_24_Mclks; };
	struct ksig
	{
		enum mod { flat, sharp } m;
		uchar n;
		enum qual { major, minor } q;
	};
	
	uchar mt;

	// These used to all be in a union but I took them out
	int seq_num;
	int ch_prefix;
	long tem;
	offs o;
	tsig t;
	ksig k;
	baws seq_spec;
	UtilStringPtr txt;
	
	enum
	{
		sequence_num   =    0,
		txt_generic    =    1,
		txt_copyright  =    2,
		txt_seq_or_trk =    3,
		txt_instrument =    4,
		txt_lyric      =    5,
		txt_marker     =    6,
		txt_cue        =    7,
		chan_prefix    = 0x20,
		trk_end        = 0x2F,
		tempo          = 0x51,
		SMPTE_offset   = 0x54,
		time_sig       = 0x58,
		key_sig        = 0x59,
		sequencer_spec = 0x7F
	};
	
	meta() { }
	meta(const meta& m) { operator=(m); }
	meta(int meta_ty, const uchar* r, long len) { init(meta_ty, r, len); }

	int sprint( char *s ) const;
	UtilStringPtr unparse() const;

	void init( int meta_ty, const uchar *r, long len );

	meta& operator=(const meta& m)
	{
		mt = m.mt;
		seq_num = m.seq_num;
		ch_prefix = m.ch_prefix;
		tem = m.tem;
		o = m.o;
		t = m.t;
		k = m.k;
		seq_spec = m.seq_spec;
		txt = m.txt;
		return (*this);
	}
};

//--------------------------------------------------------------------------------------------------

class ext_data
{
public:
	uchar st;
	baws sys_ex; // These aren't in a union anymore because they have members with constructors.
	meta meta_event; // But it doesn't take up too much more space.
	
	ext_data(const ext_data& e) { operator=(e); }
	ext_data( uchar *r, long len ) : st(smsgt::sys_ex) { sys_ex.init(r, len); }
	ext_data( int meta_ty, uchar *r, long len ) : st(smsgt::file_meta)
		{ meta_event.init(meta_ty, r, len); }
	
	long size() const { return st == smsgt::sys_ex ? sys_ex.d->length : 0; }
	
	int sprint( char *s ) const;
	UtilStringPtr unparse() const;

	ext_data& operator=(const ext_data& e)
	{
		st = e.st;
		sys_ex = e.sys_ex;
		meta_event = e.meta_event;
		return (*this);
	}
};

typedef UtilPtr<ext_data> ext_dataPtr;

//--------------------------------------------------------------------------------------------------

class MidiMsg
{
protected:
	static int LSN(uchar b) { return b & 0xF; } // least significant nibble

public:
	enum
	{
		note_off    = 0,
		note_on     = 1,
		poly_press  = 2,
		control_chn = 3,
		program_chn = 4,
		chan_press  = 5,
		pitch_bend  = 6,
		system      = 7
	};

	long time;
	uchar s, d1, d2; // 3
	ext_dataPtr ed;

	MidiMsg() { time = 0; s = 0; d1 = 0; d2 = 0; }
	MidiMsg(const MidiMsg& m) { operator=(m); }

	static int  msgt( uchar sb ) { return (sb >> 4) & 7;   } // message type (coarse distintions)
	static int smsgt( uchar sb ) { return LSN(sb); } // system message type (for system msgs)
	static int note_num(const char*); // note name -> MIDI note number

	int msgt()  const { return msgt(s); }
	int smsgt() const { return LSN(s); }
	int chan()  const { return LSN(s); }  // channel (for channel msgs)
	int dat14() const { return (int(d2) << 7) | d1; } // 14-bit data
	int bend()  const { return dat14() - 0x2000; } // pitch bend data (bipolar)
	int key()   const { return d1; }
	int vel()   const { return d2; }
	int len()   const;
	
	bool is_note_end() const { return msgt()==note_off || (msgt()==note_on && vel()==0); }
	bool is_ext() const { return msgt() == system && smsgt() == smsgt::ext; }
	
	void set_bend( float v ) { set_data14( (int) (0x2000 + v * 0x1FFF) ); } // 7
	void set_stat_byt( int msgt, int lsn ) { s = 0x80 | (msgt << 4) | lsn; } // 6
	void set_data14( int v ) { d1 = v & 0x7F; d2 = (v >> 7) & 0x7F; }       // 5
	void set_chan(int c) { s = (s & 0xf0) | c; }
	void set_key(int k) { d1 = (uchar)k; }
	void set_key(const char* k) { d1 = (uchar) note_num(k); if (d1 == 0xff) d1 = 60; }
	void set_vel(int v) { d2 = (uchar)v; }

	void set_sys_ex(              uchar *r, long len );
	void set_meta( int meta_type, uchar *r, long len );

	void sprint( char *s ) const;
	UtilStringPtr unparse() const;
	bool operator==(const MidiMsg& m) const;

	MidiMsg& operator=(const MidiMsg& m)
	{
		time = m.time;
		s = m.s;
		d1 = m.d1;
		d2 = m.d2;
		ed = m.ed;
		return (*this);
	}
};
/*
1. A MIDI_msg should never indicate end of system exclusive.  Rather, system exclusive messages 
will be assembled into a single MIDI_msg, with the end marker status byte appearing as the last 
data byte in the sys_ex_data structure pointed to.

3. The various OMS*Packet* structures use 4 bytes for data, presumably since SysEx data can be 
sent in these and anyway the structs will be aligned to at least a 16-bit boundary so it's not 
a waste of space.  Though I deal with SysEx differently, I may still use 4 bytes so 
that we can assemble 14-bit continuous controller message pairs into one message.  For now, I
will stick with 3.

5. Sets the two data bytes to contain the 14-bit quantity in the argument.

6. Sets the status byte to the provided message type and least significant nibble.

7. Takes a floating-point number in the range [-1.0...1.0] and scales it appropriately. This 
function will never produce the maximum negative pitch bend, 0x0000 (interpreted as -8192) since 
it scales symmetrically about zero and the maximum positive pitch bend is only 0xF7F7 (interpreted 
as 8191).
*/

#endif
// Author: Ben Denckla
