/*
** Copyright 1995-1996 by the Massachusetts Institute of Technology.  All
** rights reserved. For more information, see the file COPYING which came
** with this source distribution.
*/

#include "ChannelState.h"

ChannelState::ChannelState()
{
	int i;
	for (i = 0; i < 256; i++)
		notes[i] = 0; // all notes are off
	for (i = 0; i < 32; i++)
		control14[i] = 0;
	for (i = 0; i < 16; i++)
		control7[i] = 0;
	control14[7] = 16383; // volume at maximum
	control14[8] = 64*128; // balance at middle
	control14[10] = control14[8]; // pan at middle
	program = 0; // piano
	channelPressure = 0;
	pitchBend = 64*128; // no bend
}

void ChannelState::update(const MIDI_msg* m)
{
	switch (m->msgt())
	{
	case MIDI_msg::note_off:
		notes[m->key()]--;
		break;
	case MIDI_msg::note_on:
		if (m->vel() == 0) // a zero-velocity "note on" is a note end
			notes[m->key()]--;
		else
			notes[m->key()]++;
		break;
	case MIDI_msg::poly_press:
		break; // we don't keep track of polyphonic pressure
	case MIDI_msg::control_chn:
		if (m->d1 < 32) // MSB of 14-bit control change
			control14[m->d1] = m->d2 * 128; // reset LSB to zero
		else if (m->d1 < 64) // LSB of 14-bit control change
			control14[m->d1-32] = (control14[m->d1-32] & 0xf8) | m->d2; // preserve MSB
		else if (m->d1 < 96) // 7-bit control change
			control7[m->d1-64] = m->d2;
		break;
	case MIDI_msg::program_chn:
		program = m->d1;
		break;
	case MIDI_msg::chan_press:
		channelPressure = m->d2;
		break;
	case MIDI_msg::pitch_bend:
		pitchBend = m->dat14();
		break;
	default:
		break;
	}
}

void ChannelState::change(const ChannelState* s, ioref n, int c, void (*outfunc)(const MIDI_msg&))
{
	MIDI_msg m;
	m.node = n;
	int i;
	for (i = 0; i < 256; i++) // turn off notes
		if (notes[i] < s->notes[i])
		{
			m.set_stat_byt(MIDI_msg::note_on, c);
			m.set_key(i); // this may be out of range, but OutputUnit will wrap it
			m.set_vel(0); // a zero-velocity "note on" is a note end
			for (int j = 0; j < (s->notes[i] - notes[i]); j++)
				outfunc(m); // repeat appropriate number of times
		}
	for (i = 0; i < 32; i++) // fix 14-bit controls
		if (control14[i] != s->control14[i])
		{
			m.set_stat_byt(MIDI_msg::control_chn, c);
			m.d1 = i;
			m.d2 = control14[i] >> 7;
			outfunc(m); // send MSB
			m.d1 += 32;
			m.d2 = control14[i] & 0x7f;
			outfunc(m); // send LSB
		}
	for (i = 0; i < 16; i++) // fix 7-bit controls
		if (control7[i] != s->control7[i])
		{
			m.set_stat_byt(MIDI_msg::control_chn, c);
			m.d1 = i + 64;
			m.d2 = control7[i];
			outfunc(m);
		}
	if (program != s->program)
	{
		m.set_stat_byt(MIDI_msg::program_chn, c);
		m.d1 = program;
		outfunc(m);
	}
	if (channelPressure != s->channelPressure)
	{
		m.set_stat_byt(MIDI_msg::chan_press, c);
		m.d1 = channelPressure;
		outfunc(m);
	}
	if (pitchBend != s->pitchBend)
	{
		m.set_stat_byt(MIDI_msg::pitch_bend, c);
		m.set_data14(pitchBend);
		outfunc(m);
	}
	for (i = 0; i < 256; i++) // turn on notes
		if (notes[i] > s->notes[i])
		{
			m.set_stat_byt(MIDI_msg::note_on, c);
			m.set_key(i); // this may be out of range, but OutputUnit will wrap it
			m.set_vel(32); // a nice, soft velocity (since we're just guessing)
			for (int j = 0; j < (notes[i] - s->notes[i]); j++)
				outfunc(m); // repeat appropriate number of times
		}
}

// Author: Patrick Pelletier
