/*
** Copyright 1995-1996 by the Massachusetts Institute of Technology.  All
** rights reserved. For more information, see the file COPYING which came
** with this source distribution.
*/

#include "ScorePlayer.h"
#include "data_structs.h"
#include "BinStatistic.h"

typedef UtilPtr<State> StatePtr;

void ScorePlayer::defaultOutFunc(const MIDI_msg& m)
{
	oap->write(&m);
}

static int binner(int i)
{
	return (i*5);
}

StatePtr ScorePlayer::defaultState;

static void iterator(ioref, int, bool* e, void*)
{
	*e = 0;
}

ScorePlayer::ScorePlayer(ScorePtr sp, void (*func)(const MIDI_msg&))
{
	score = sp;
	outfunc = func;
	u = score->getChannels();
	state = new State(u);
	playing = 0;
	muted = 0;
	looped = 0;
	scale = (ScoreScale) 1;
	if (defaultState == NULL)
	{
		Score::ChannelUseMapPtr d = new Score::ChannelUseMap;
		d->iterate(iterator, NULL);
		defaultState = new State(d);
	}
	leftover = -1;
	index = 0;
	filters = new MidiFilterPtrArray(0);
	functask = NULL;
	funcnum = -1;
	calling_func = false;
}

ScorePlayer::~ScorePlayer()
{
	stop();
}

void ScorePlayer::startMultiple(ScorePlayerPtrArray* a)
{
	ScheduleTime t = oap->time();
	int i;
	for (i = 0; i < a->length; i++)
	{
		ScorePlayer* sp = (*a)[i];
		if (sp->playing || (sp->score->getNumEvents() == 0))
			continue;
		if (!(sp->muted))
			sp->state->change(defaultState, sp->outfunc);
		sp->do_start(t);
		sp->playing = 1;
	}
}

void ScorePlayer::start()
{
//	for (int i = 0; i < score->getNumEvents(); i++)
//		tw.spr("time of %d is %ld\r", i, (*score)[i].time);
	if (playing || (score->getNumEvents() == 0))
		return;
	if (!muted)
		state->change(defaultState, outfunc);
	ScheduleTime t = oap->time();
	do_start(t);
	playing = 1;
}

void ScorePlayer::do_start(ScheduleTime tt)
{
	ScheduleTime t = tt;
	lastWakeUp = t;
	if (leftover != -1 && index > 0)
		t += (ScheduleTime) (scale*((*score)[index].time-((*score)[index-1].time+leftover)));
	nextWakeUp = t;
	if (nextWakeUp < lastWakeUp)
	{
		tw.pr("We scheduled %ld demimilliseconds in the past\r",lastWakeUp-nextWakeUp);
//		nextWakeUp = lastWakeUp;
//		leftover = -1;
	}
	task_data td(callback, (void*) this, nextWakeUp);
	if (schedp->insert(td, (task**)(&ott)) != no_err)
	{
		if (!muted)
			defaultState->change(state, outfunc);
		ew.w("schedp->insert failed");
		leftover = -1;
		playing = 0;
		return;
	}
	if (!calling_func)
	{
		if (funcnum != -1)
		{
			ew.w("something is wrong in ScorePlayer::do_start");
			return;
		}
		ScheduleTime nextfunc = decide_next_function();
		if (nextfunc != NEVER)
		{
			task_data func_td(func_callback, (void*) this, nextfunc+tt);
			if (schedp->insert(func_td, &functask) != no_err)
				ew.w("schedp->insert failed for functask in ScorePlayer::do_start");
		}
	}
}

void ScorePlayer::stopMultiple(ScorePlayerPtrArray* a)
{
	ScheduleTime t = oap->time();
	int i;
	for (i = 0; i < a->length; i++)
	{
		ScorePlayer* sp = (*a)[i];
		if (!(sp->playing))
			continue;
		if (sp->do_stop(t) == no_err)
		{
			if (!(sp->muted))
				defaultState->change(sp->state, sp->outfunc);
			sp->playing = 0;
		}
	}
}

void ScorePlayer::stop()
{
	if (!playing)
		return;
	ScheduleTime t = oap->time();
	if (do_stop(t) == no_err)
	{
		if (!muted)
			defaultState->change(state, outfunc);
		playing = 0;
	}
}

err_ind ScorePlayer::do_stop(ScheduleTime t)
{
	if (schedp->cancel((task*)ott) != no_err)
	{
		ew.w("schedp->cancel failed\r");
		return err;
	}
	if (leftover == -1)
		leftover = 0;
	leftover += (ScoreTime) ((t-lastWakeUp)/scale);
	if ((!calling_func) && funcnum != -1 && schedp->cancel(functask) != err)
	{
		functask = NULL;
		funcnum = -1;
	}
	return no_err;
}

typedef Score::ChannelUseMapPtr ChannelUseMapPtr;

StatePtr ScorePlayer::calculate_state(ChannelUseMapPtr u, ScorePtr sc, int i)
{
	StatePtr result = new State(u);
	if (i > sc->getNumEvents())
		i = sc->getNumEvents();
	int j = 0;
	while (j < i)
		result->update(&((*sc)[j++]));
	return result;
}

void ScorePlayer::setIndex(int i)
{
	ScheduleTime t = oap->time();
	if (playing)
		if (do_stop(t) == err)
			return;
	StatePtr newState = calculate_state(u, score, i);
	if (playing && !muted)
		newState->change(state, outfunc);
	state = newState;
	index = i;
	leftover = -1;
	if (playing)
		do_start(t);
}

void ScorePlayer::setTimeUnscaled(ScoreTime x)
{
	ScheduleTime t = oap->time();
	if (playing)
		if (do_stop(t) == err)
			return;
	index = 0;
	leftover = -1;
	if (score->getNumEvents() > 0)
		for (int i = 0; i < score->getNumEvents(); i++)
			if ((*score)[i].time >= x)
			{
				if (i == 0 || (*score)[i].time == x)
					leftover = -1;
				else
					leftover = (x - (*score)[i-1].time);
				index = i;
				break;
			}
	StatePtr newState = calculate_state(u, score, index);
	if (playing && !muted)
		newState->change(state, outfunc);
	state = newState;
	if (playing)
		do_start(t);
}

ScoreTime ScorePlayer::getTimeUnscaled() const
{
	ScheduleTime t = oap->time();
	ScoreTime lo = leftover;
	if (lo == -1)
		lo = 0;
	if (playing)
		lo += (ScoreTime) ((t-lastWakeUp)/scale + 0.5);
	ScoreTime base = ((index > 0) ? (*score)[index-1].time : 0);
	return base+lo;
}

void ScorePlayer::setScale(ScoreScale s)
{
	ScheduleTime t = oap->time();
	if (playing)
		if (do_stop(t) == err)
			return;
	scale = s;
	if (playing)
		do_start(t);
}

void ScorePlayer::setMute(bool mu)
{
	mu = ((mu == 1) ? 1 : 0);
	if (mu == muted)
		return;
	ScheduleTime t = oap->time();
	if (playing && do_stop(t) == err)
		return;
	muted = mu;
	if (muted)
		defaultState->change(state, outfunc);
	else
		state->change(defaultState, outfunc);
	if (playing)
		do_start(t);
}

ScheduleTime ScorePlayer::func_callback(void* v)
{
	ScorePlayer* p = (ScorePlayer*) v;
	if (p->funcnum < 0 || p->funcnum > p->score->funcs->length)
	{
		tw.spr("Something is wrong in ScorePlayer::func_callback()\r");
		return NEVER;
	}
	p->calling_func = true;
//	tw.spr("Calling function from %p\r",p);
	((*(p->score->funcs))[p->funcnum])((*(p->score->args))[p->funcnum], p); // call the function
	p->calling_func = false;
	return p->decide_next_function_for_callback();
}

ScheduleTime ScorePlayer::callback(void* dwUser)
{
	ScorePlayer* p = (ScorePlayer*) dwUser;
	MIDI_msg m = ((*(p->score))[p->index]);
	if (p->filters->length > 0)
		for (int i = 0; i < p->filters->length; i++)
			(*(p->filters))[i]->filter(&m);
	if (!(p->muted))
		p->outfunc(m);
	p->state->update(&m);
	p->index++;
	if (p->index >= p->score->getNumEvents())
	{
		p->index = 0;
		p->leftover = -1;
		if (!(p->muted))
			defaultState->change(p->state, (p->outfunc));
		if (!(p->looped))
		{
			p->playing = 0;
			return NEVER;
		}
		else
			m.time = (*(p->score))[0].time;
	}
	p->lastWakeUp = p->nextWakeUp;
//	tw.spr("m.time = %ld m.s = %d m.d1 = %d m.d2 = %d\r", m.time, m.s, m.d1, m.d2);
//	tw.spr("((*(p->score))[p->index]).time = %ld, m.time = %ld p->scale = %f\r",((*(p->score))[p->index]).time, m.time, p->scale);
	p->nextWakeUp += (ScheduleTime) (((((*(p->score))[p->index]).time - m.time) * p->scale) + 0.5);
//	tw.spr("We added %ld to nextWakeUp\r",p->nextWakeUp - p->lastWakeUp);
	p->leftover = -1;
	return (p->nextWakeUp - p->lastWakeUp);
}

int ScorePlayer::find_filter(const char* f) const
{
	if (filters->length > 0)
		for (int i = 0; i < filters->length; i++)
			if (strcmp((*filters)[i]->getIdentity(), f) == 0)
				return i;
	return -1;
}

static void iterator(ioref n, int c, ChannelStatePtr* e, void* x)
{
	if (*e == NULL)
		return;
	State* s = (State*)x;
	**e = *((*s)(n,c));
}

void ScorePlayer::addFilter(MidiFilterPtr p)
{
	StatePtr newState = new State(u);
	ScheduleTime t = oap->time();
	if (playing)
		if (do_stop(t) == err)
			return;
	newState->iterate(iterator, (void*)state);
	int i = find_filter(p->getIdentity());
	if (i == -1)
	{
		MidiFilterPtrArrayPtr temp = new MidiFilterPtrArray(filters->length +1);
		(*temp) = (*filters);
		i = filters->length;
		filters = temp;
	}
	else
		(*filters)[i]->unapply(newState);
	(*filters)[i] = p;
	p->apply(newState);
	if (playing && !muted)
		newState->change(state, outfunc);
	state = newState;
	if (playing)
		do_start(t);
}

void ScorePlayer::removeFilter(const char* f)
{
	int i = find_filter(f);
	if (i == -1)
		return;
	StatePtr newState = new State(u);
	ScheduleTime t = oap->time();
	if (playing)
		if (do_stop(t) == err)
			return;
	newState->iterate(iterator, (void*)state);
	(*filters)[i]->unapply(newState);
	MidiFilterPtrArrayPtr temp = new MidiFilterPtrArray(filters->length-1);
	int k = 0;
	if (filters->length > 0)
		for (int j = 0; j < filters->length; j++)
			if (j != i)
				(*temp)[k++] = (*filters)[j];
	if (playing && !muted)
		newState->change(state, outfunc);
	state = newState;
	filters = temp;
	if (playing)
		do_start(t);
}

typedef UtilPtr<MidiFilter> MidiFilterPtr;

MidiFilterPtr ScorePlayer::getFilter(const char* f) const
{
	int i = find_filter(f);
	if (i == -1)
		return NULL;
	return ((*filters)[i]);
}

ScheduleTime ScorePlayer::decide_next_function()
{
	ScoreTime lo = leftover;
	if (lo == -1)
		lo = 0;
	ScoreTime base = ((index > 0) ? (*score)[index-1].time : 0);
	ScoreTime pos =  base+lo;
	int i;
	for (i = 0; i < score->times->length; i++)
	{
		if ((*(score->times))[i] >= pos)
		{
			funcnum = i;
			return ((*(score->times))[i] - pos);
		}
	}
	funcnum = -1;
	return NEVER;
}

ScheduleTime ScorePlayer::decide_next_function_for_callback()
{
	funcnum++;
	if (funcnum >= score->times->length)
	{
		funcnum = -1;
		return NEVER;
	}
	return ((*score->times)[funcnum]-(*score->times)[funcnum-1]);
}

// Author: Patrick Pelletier
